#include "CloseUp.h"
#include <fstream>
#include <sys/time.h>
#include <unistd.h>
#include "opencv2/opencv.hpp"
#include "dynamixel_sdk/dynamixel_sdk.h"
#include <cmath>


using namespace cv;
using namespace std;




void AssigneInputsDynamixel (int OffsetAlpha, int OffsetBeta, int RangesAlpha[2], int RangesBeta[2],Point3f CloseUpPosition)
{
	alphaOffset=OffsetAlpha;
	betaOffset=OffsetBeta;
	
	dOriginCloseup_x=CloseUpPosition.x;
	dOriginCloseup_y=CloseUpPosition.y;
	dOriginCloseup_z=CloseUpPosition.z;
	
	alphaRange[0]=RangesAlpha[0];
	betaRange[0]=RangesBeta[0];
	alphaRange[1]=RangesAlpha[1];
	betaRange[1]=RangesBeta[1];
	
}


void connectionDynamixel(){

    int dxl_comm_result = COMM_TX_FAIL;             // Communication result
    uint8_t dxl_error = 0;                          // Dynamixel error

    // Open port
    if (portHandler->openPort())
        printf("SudObjectCloseupeeded to open the port!\n");
    else
        printf("Failed to open the port!\n");

    // Set port baudrate
    if (portHandler->setBaudRate(BAUDRATE))
        printf("SudObjectCloseupeeded to change the baudrate!\n");
    else
        printf("Failed to change the baudrate!\n");

    // Enable Dynamixel#1 Torque
    dxl_comm_result = packetHandler->write1ByteTxRx(portHandler, DXL1_ID, ADDR_MX_TORQUE_ENABLE, TORQUE_ENABLE, &dxl_error);
    if (dxl_comm_result != COMM_SUCCESS)
        packetHandler->getTxRxResult(dxl_comm_result);
    else if (dxl_error != 0)
        packetHandler->getRxPacketError(dxl_error);
    else
        printf("Dynamixel#%d has been connected \n", DXL1_ID);

    // Enable Dynamixel#2 Torque
    dxl_comm_result = packetHandler->write1ByteTxRx(portHandler, DXL2_ID, ADDR_MX_TORQUE_ENABLE, TORQUE_ENABLE, &dxl_error);
    if (dxl_comm_result != COMM_SUCCESS)
        packetHandler->getTxRxResult(dxl_comm_result);
    else if (dxl_error != 0)
        packetHandler->getRxPacketError(dxl_error);
    else
        printf("Dynamixel#%d has been connected \n", DXL2_ID);

}



/*
 * setDynamixelMotors
 * Set the Dynamixel motors in the correct position, depending on the objects position
 */
void setDynamixelMotors(Point3f point3DReal){


	dObjectCloseup_x = point3DReal.x-dOriginCloseup_x;
    dObjectCloseup_y = point3DReal.y-dOriginCloseup_y;
    dObjectCloseup_z = point3DReal.z-dOriginCloseup_z;

    dObjectCloseup_xy = pow(dObjectCloseup_x, 2) + pow(dObjectCloseup_y, 2);
    dObjectCloseup_xy = sqrt(dObjectCloseup_xy);



	
    // angle of the first motor
    float alpha;
    int alphaMotor;

    //angle of the second motor
    float beta;
    int betaMotor;

    // addParam result
    bool dxl_addparam_result = false;
    // Communication result
    int dxl_comm_result = COMM_TX_FAIL;

    // Motors goal position
    uint8_t param_goal_position[4];

    // calculate the angle of the first motor
    alpha = atan(dObjectCloseup_x/dObjectCloseup_y);
    alpha = alpha*(180/M_PI);           //convert in degrees
    alpha = alpha*(4095/360);           //convert in motor step

    alphaMotor=-int(alpha)+alphaOffset;//add the offset over z axis

    // calculate the angle of the second motor
    beta=atan(dObjectCloseup_z/dObjectCloseup_xy);
    beta=beta*(180/M_PI);               //convert in degrees
    beta=beta*(4095/360);               //convert in motor step

    betaMotor=int(beta)+betaOffset;// the offset over the y axis

    // Allocate goal position value into byte array
    param_goal_position[0] = DXL_LOBYTE(DXL_LOWORD(alphaMotor));
    param_goal_position[1] = DXL_HIBYTE(DXL_LOWORD(alphaMotor));
    param_goal_position[2] = DXL_LOBYTE(DXL_HIWORD(alphaMotor));
    param_goal_position[3] = DXL_HIBYTE(DXL_HIWORD(alphaMotor));

    // Add Dynamixel#1 goal position value to the Syncwrite storage
    dxl_addparam_result = groupSyncWrite.addParam(DXL1_ID, param_goal_position);

    if (dxl_addparam_result != true){
      fprintf(stderr, "[ID:%03d] groupSyncRead addparam failed", DXL1_ID);
    }

    // Allocate goal position value into byte array
    param_goal_position[0] = DXL_LOBYTE(DXL_LOWORD(betaMotor));
    param_goal_position[1] = DXL_HIBYTE(DXL_LOWORD(betaMotor));
    param_goal_position[2] = DXL_LOBYTE(DXL_HIWORD(betaMotor));
    param_goal_position[3] = DXL_HIBYTE(DXL_HIWORD(betaMotor));

    // Add Dynamixel#2 goal position value to the Syncwrite storage
    dxl_addparam_result = groupSyncWrite.addParam(DXL2_ID, param_goal_position);

    if (dxl_addparam_result != true){
      fprintf(stderr, "[ID:%03d] groupSyncRead addparam failed", DXL2_ID);
    }

    // Syncwrite goal position
    if((alphaMotor<alphaRange[1] & alphaMotor>alphaRange[0]) & (betaMotor<betaRange[1] & betaMotor>betaRange[0]) )
        dxl_comm_result = groupSyncWrite.txPacket();

    if (dxl_comm_result != COMM_SUCCESS)
        printf("%s\n", packetHandler->getTxRxResult(dxl_comm_result));

    // Clear syncwrite parameter storage
    groupSyncWrite.clearParam();
}


