/*
 Ali NouriZonoz  CMU Geneve
 July 2020
 */

#include "opencv2/opencv.hpp"
#include <iomanip>
#include <unistd.h>
#include <arpa/inet.h>
#include <fstream>
#include <atomic>
#include <sys/socket.h>
#include "../QtFiles/LoadConfiguration.h"
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <fstream>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <stdint.h>
#include <netinet/in.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>


using namespace cv;
using namespace std;

extern void connectionDynamixel();
extern void AssigneInputsDynamixel (int OffsetAlpha, int OffsetBeta, int RangesAlpha[2], int RangesBeta[2],Point3f CloseUpPosition);
extern void setDynamixelMotors(Point3f point3DReal);
extern int Triangulate(const vector<Mat> PM, const vector<int> index, const vector<Point2f> projPoints, Mat& X);

//********************************************************************************************* Defining Global Variables ******************************************************************************************//
//******************************************************************************************************************************************************************************************************************//
//******************************************************************************************************************************************************************************************************************//

int CameraNumbers;
vector<Point2f> POS;
vector<int> PORT;
atomic <bool> TrackDone;
int UseLiveCamera=0;
int UseRecoBox=0;
int RecoBoxMode=0;
int UseCloseUp=0;
vector<Mat> projMatrix;
Mat RWTransition;
Mat matrixCam   = (Mat_<float>(3,3) << 4.6996008656064788e+02, 0, 3.1550000000000000e+02, 0,  4.6996008656064788e+02, 2.5350000000000000e+02, 0, 0, 1);
Mat distCoeffs  = (Mat_<float>(1,4) << -3.2842997314171629e-01, 1.6718837850893636e-01, 0, 0);
string OutDir;
ofstream OutPutTrajectory;

LiveCamera LCam;
ForagingRecoBox FReco;
ConditioningRecoBox CReco;
CloseUpSystem CloseUp;
int DoPlot=0;
int LiveFrameNum=0;
int CloseUpFramne=0;
Point3f POS3D;
int BehaviorExecuted=0;
int FrameNumber=0;
int RewardNum=0;
int StimulusNum=0;
Mat CloseUpIMGRes,LiveIMGRes;
struct timeval TBegin, TPosition, TRecoBox,TLiveCap,TCloseCap;

//********************************************************************************************* Functions******************************************************************************//
//*****************************************************************************************************************************************************************************************//
//****************************************************************************************************************************************************************************************//

/* getSinglePointCameras
 * get the 2D coordinates (projPoints) and the index of the camera seeing the object
 * only the 2D coordinates where the camera sees the object are taken into account
 */
void getSinglePointCameras(vector<Point2f>& projPoints, vector<int>& index, Mat matrixCam, Mat distCoeffs,int CameraNumber){

    //Vectors for single points inside the Triangulation loop - for Onlien tracking
    vector<Point2f> obj(CameraNumber);

    projPoints.clear();
    index.clear();

    for(int i=0; i<CameraNumber; i++)
        obj[i] = POS[i];

    // undistort the 2D points using the corresponding coefficients
    vector<Point2f> undist_obj(CameraNumber);
    undistortPoints(obj, undist_obj, matrixCam, distCoeffs);

    // get the 2D coordinates of the projection points in pixel
    vector<Point2f> pixel_obj(CameraNumber);
    for(int i=0; i<CameraNumber; i++){
        pixel_obj[i].x = undist_obj[i].x * matrixCam.at<float>(0, 0) + matrixCam.at<float>(0, 2);
        pixel_obj[i].y = undist_obj[i].y * matrixCam.at<float>(1, 1) + matrixCam.at<float>(1, 2);
    }
    // if the coordinates are (0,0), do not take value
    for (int i=0; i<CameraNumber; i++){
        if(obj[i] != Point2f(0,0)){
            index.push_back(i);
            projPoints.push_back(pixel_obj[i]);
        }
    }
}

//calculate 3D distance between 2 points

float calcdist(cv::Point3f A, cv::Point3f B)
{
	float Dist=float ( sqrt (  (A.x-B.x)*(A.x-B.x) +  (A.y-B.y)*(A.y-B.y) + (A.z-B.z)*(A.z-B.z)     ) );
	return Dist;
}



/* Thread function
 * One thread per camera in order to get the object 2D projections coordinates sent by the Jetsons via UDP
 * One thread to plot the triangulated 3D point
 * 
 */
void *Track(void *threadid)
{
    long tid;
    tid = (long)threadid;
    
    
    
    if(tid==0)
    {
    	//3Dtracking
    	vector<Point2f> projPoints;
   		Mat point3D,point3DReal;
    	vector<int> index;
    	while(1){
    		usleep(5000);
			getSinglePointCameras(projPoints, index, matrixCam, distCoeffs,CameraNumbers);
		    int r = Triangulate(projMatrix, index,projPoints,point3D);
		    if(r){
		    	point3DReal = RWTransition*point3D;
		    	POS3D=Point3f( point3DReal.at<float>(0, 0),point3DReal.at<float>(0, 1),point3DReal.at<float>(0, 2));
		        OutPutTrajectory << point3DReal.at<float>(0,0)<<" "<<point3DReal.at<float>(0,1)<<" "<<point3DReal.at<float>(0,2)<<" ";
		        if(UseLiveCamera==1)
		        	OutPutTrajectory<<LiveFrameNum<<" ";
		        if(UseCloseUp==1)
		        	OutPutTrajectory<<CloseUpFramne<<" ";
		        if(UseRecoBox==1)
		        	OutPutTrajectory<<StimulusNum<<" "<<RewardNum<<" ";
		       	if(StimulusNum>0)
		       		StimulusNum=0;
		       	if(RewardNum>0)
		       		RewardNum=0;
		        
		        OutPutTrajectory<<endl;
		        
		       
		        DoPlot=1;    
		    }
		}
	}
    
    
    if(tid==1)
	{
		//gnuplot
		string PlottingCommand= "../src/RunEthoLoop/Plot.sh "+OutDir+" -e";
		while(!TrackDone)
			if (DoPlot==1)
				system(PlottingCommand.c_str());
		pthread_exit(NULL);
	}
	
	
	
	if(tid==2 & UseLiveCamera==1)
	{
		//LiveCamera
		cout<<"Live"<<endl;
		sleep(4);
		Mat RAW_IMG,Col_IMG;
			
		VideoCapture LiveCap(CV_CAP_XIAPI+0);
		LiveCap.set(CV_CAP_PROP_XI_AEAG,0);
       	LiveCap.set(CV_CAP_PROP_XI_EXPOSURE,LCam.Exposure);
       	LiveCap.set(CV_CAP_PROP_XI_GAIN,LCam.Gain);
		LiveCap.set(CV_CAP_PROP_XI_DATA_FORMAT,5);
		Size Live_Resize(LCam.Width,LCam.Height);
		int H=LiveCap.get(4);
		int W=LiveCap.get(3);

		VideoWriter video(OutDir+"/Live.avi",CV_FOURCC('M','P','E','G'),60,Size(W,H),true);
		while(!TrackDone)
		{
			LiveCap>>RAW_IMG;
			LiveFrameNum=LiveFrameNum+1;	
			cvtColor(RAW_IMG,Col_IMG,COLOR_BayerBG2BGR);
			video.write(Col_IMG);
			resize(Col_IMG,LiveIMGRes,Live_Resize);
			if(UseLiveCamera==1){
				imshow("LIVE",LiveIMGRes);
				if(UseCloseUp==1)
					imshow("CloseUp",CloseUpIMGRes);
				waitKey(1);
			}
		}	
		pthread_exit(NULL);
	} 
	
	
	
	
	if(tid==3 & UseRecoBox==1)
	{
	
		
		//RECO Box		
		if(RecoBoxMode==0)
		{
			string Command=CReco.Stimulus;
			string RewCommand="R"+to_string(CReco.RewardSize);
			int numbytes;
			struct timeval T1,T2;
			//create socket
			int sockfd_ard,rv_ard,numbytes_ard;
			struct addrinfo hints_ard, *servinfo_ard, *p_ard;
			memset(&hints_ard, 0, sizeof hints_ard);
			hints_ard.ai_family = AF_UNSPEC;
			hints_ard.ai_socktype = SOCK_DGRAM;
			if ((rv_ard = getaddrinfo(CReco.IP.c_str(), CReco.Port.c_str(), &hints_ard, &servinfo_ard)) != 0) 
				fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv_ard));	
			for(p_ard = servinfo_ard; p_ard != NULL; p_ard = p_ard->ai_next) {
				if ((sockfd_ard = socket(p_ard->ai_family, p_ard->ai_socktype,p_ard->ai_protocol)) == -1) {
					perror("talker: socket");
					continue;
				}
				break;
			}
			
			int ActivateReward;
			while(!TrackDone)
			{
				ActivateReward=0;		
				if(calcdist(POS3D,CReco.TargetPos)<CReco.Target_Zone)
				{
					int NumTimesClicked=0;
					if(CReco.ConditioningType==0){
						
						numbytes = sendto(sockfd_ard, Command.c_str(), (Command.size()), 0, p_ard->ai_addr, p_ard->ai_addrlen);
						StimulusNum=1;
						ActivateReward=1;
					}
					if(CReco.ConditioningType==1 & BehaviorExecuted==1)
					{
						numbytes = sendto(sockfd_ard, Command.c_str(), (Command.size()), 0, p_ard->ai_addr, p_ard->ai_addrlen);
						StimulusNum=1;
						ActivateReward=1;
					}
					gettimeofday(&T1,NULL);
					while(ActivateReward==1)
					{	
	
						if(calcdist(POS3D,CReco.SelfPos)<CReco.RECO_Zone)
						{
							numbytes = sendto(sockfd_ard, RewCommand.c_str(), (RewCommand.size()), 0, p_ard->ai_addr, p_ard->ai_addrlen);
							RewardNum=1;
							break;
						}
						gettimeofday(&T2,NULL);
						unsigned long TimePassed= 1000000*(T2.tv_sec-T1.tv_sec)+(T2.tv_usec-T1.tv_usec);
						TimePassed=TimePassed/1000;
						cout<<TimePassed<<" "<<CReco.Repeat<<" "<<NumTimesClicked<<endl;
						if(TimePassed>CReco.Timeout*1000)
							break;
						if(TimePassed/(CReco.Repeat*1000)>NumTimesClicked){
							NumTimesClicked=NumTimesClicked+1;
							numbytes = sendto(sockfd_ard, Command.c_str(), (Command.size()), 0, p_ard->ai_addr, p_ard->ai_addrlen);	
						}		
					}	
				}
			}
		}
		
		






		
		if(RecoBoxMode==1)
		{
			struct timeval T1,T2;
			vector <int> RecoNum;
			std::stringstream SeqForage(FReco.ForagingSeq);
			int SeqLen=0;
			int index=-1;
    		int ChosenReco;
			
			
			if(FReco.ForagingType==0){
				for(int i = 0;i<FReco.ForagingSeq.size(); i++) 
					if(FReco.ForagingSeq[i]>='0' && FReco.ForagingSeq[i]<='9') 
		   		 		SeqLen++;
				
				for(int i=0;i<SeqLen;i++)
				{
					int Num;
					SeqForage>>Num;
					RecoNum.push_back(Num);
				}
    		}
    		
    		
			while(!TrackDone)
			{
				if(FReco.ForagingType==0){
					index=index+1;
					if(index==RecoNum.size())
						index=0;	
					ChosenReco=RecoNum[index];
				}
				if(FReco.ForagingType==1)
					ChosenReco=int(rand()%(FReco.RecoNums))+1;
					
				
				string Command=FReco.Stimulus[ChosenReco-1];
				string RewCommand="R"+to_string(FReco.RewardSize[ChosenReco-1]);
				int numbytes;
				
				//create socket
				int sockfd_ard,rv_ard,numbytes_ard;
				struct addrinfo hints_ard, *servinfo_ard, *p_ard;
				memset(&hints_ard, 0, sizeof hints_ard);
				hints_ard.ai_family = AF_UNSPEC;
				hints_ard.ai_socktype = SOCK_DGRAM;
				if ((rv_ard = getaddrinfo(FReco.IP[ChosenReco-1].c_str(), FReco.Port.c_str(), &hints_ard, &servinfo_ard)) != 0)
					fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv_ard));	
				for(p_ard = servinfo_ard; p_ard != NULL; p_ard = p_ard->ai_next) {
					if ((sockfd_ard = socket(p_ard->ai_family, p_ard->ai_socktype,p_ard->ai_protocol)) == -1) {
						perror("talker: socket");
						continue;
					}
					break;
				}
				
				if(calcdist(POS3D,FReco.SelfPos[ChosenReco-1])>FReco.RECO_Zone)
				{
					int NumTimesClicked=0;
					gettimeofday(&T1,NULL);
					numbytes = sendto(sockfd_ard,  Command.c_str(), (Command.size()), 0, p_ard->ai_addr, p_ard->ai_addrlen);
					StimulusNum=ChosenReco;
					while(1){
						gettimeofday(&T2,NULL);
						if(calcdist(POS3D,FReco.SelfPos[ChosenReco-1])<=FReco.RECO_Zone){
							numbytes = sendto(sockfd_ard,  RewCommand.c_str(), (RewCommand.size()), 0, p_ard->ai_addr, p_ard->ai_addrlen);
							RewardNum=ChosenReco;
							break;
						}
						unsigned long TimePassed= 1000000*(T2.tv_sec-T1.tv_sec)+(T2.tv_usec-T1.tv_usec);
						TimePassed=TimePassed/1000;
						if(TimePassed>FReco.Timeout*1000)
							break;
						if(TimePassed/(FReco.Repeat*1000)>NumTimesClicked){
							NumTimesClicked=NumTimesClicked+1;
							numbytes = sendto(sockfd_ard,  Command.c_str(), (Command.size()), 0, p_ard->ai_addr, p_ard->ai_addrlen);				
						}	
					}				
				}
			}	
		}
		
		pthread_exit(NULL);
	}
	
	
	
	if(tid==4 & UseRecoBox==1 & RecoBoxMode==0 & CReco.ConditioningType==1)
	{
		struct sockaddr_in serveraddr; /* server's addr */
		struct sockaddr_in clientaddr; /* client addr */
		int sockfd; /* socket */
		int portno; /* port to listen on */
		socklen_t clientlen;
		int optval; /* flag value for setsockopt */
		int n; /* message byte size */
		char buf[1024];
		optval = 1;
		sockfd = socket(AF_INET, SOCK_DGRAM, 0);
		portno = atoi(CReco.BehaviourPort.c_str());
		setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,(const void *)&optval , sizeof(int));
		bzero((char *) &serveraddr, sizeof(serveraddr));
		serveraddr.sin_family = AF_INET;
		serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
		serveraddr.sin_port = htons((unsigned short)portno);
		int b=bind(sockfd, (struct sockaddr *) &serveraddr, sizeof(serveraddr));
		clientlen = sizeof(clientaddr);	
		while(!TrackDone)
		{
			n = recvfrom(sockfd, buf, 1024, 0,(struct sockaddr *) &clientaddr, &clientlen);
			BehaviorExecuted=buf[0];
			FrameNumber=buf[1];
		}	
		pthread_exit(NULL);
	} 
	
	
	
	
	if(tid==5 & UseCloseUp==1)
	{		
		connectionDynamixel();
		AssigneInputsDynamixel (CloseUp.OffsetY,CloseUp.OffsetZ, CloseUp.MinMaxY, CloseUp.MinMaxZ,CloseUp.CL_Position);
		while(!DoPlot){}
		while(!TrackDone)
		{
			setDynamixelMotors(POS3D);
		}
		pthread_exit(NULL);
	}
	
	
		
	if(tid==6 & UseCloseUp==1)
	{		
		//LiveCamera
		cout<<"CloseUp Camera"<<endl;
		sleep(4);
		Mat RAW_IMG;
		VideoCapture CloseUpCap(CV_CAP_XIAPI+1);
		CloseUpCap.set(CV_CAP_PROP_XI_AEAG,0);
       	CloseUpCap.set(CV_CAP_PROP_XI_EXPOSURE,CloseUp.Exposure);
       	CloseUpCap.set(CV_CAP_PROP_XI_GAIN,CloseUp.Gain);
		CloseUpCap.set(CV_CAP_PROP_XI_DATA_FORMAT,5);
		Size CloseUp_Resize(CloseUp.Width,CloseUp.Height);
		int H=CloseUpCap.get(4); 
		int W=CloseUpCap.get(3);

		VideoWriter video(OutDir+"/CloseUp.avi",CV_FOURCC('M','P','E','G'),60,Size(W,H),true);
		while(!TrackDone)
		{
			CloseUpCap>>RAW_IMG;
			CloseUpFramne=CloseUpFramne+1;	
			video.write(RAW_IMG);
			resize(RAW_IMG,CloseUpIMGRes,CloseUp_Resize);
			if(UseLiveCamera==0){
				imshow("CloseUp",CloseUpIMGRes);
				waitKey(1);
			}
				
		}	
		pthread_exit(NULL);
	} 
	
	                             
    if(tid >6 ){
    
    	int BUFSIZE=30;
        struct sockaddr_in serveraddr;  /* server's addr */
        struct sockaddr_in clientaddr;  /* client addr */
        struct hostent *hostp;          /* client host info */
        int sockfd;                     /* socket */
        int portno;                     /* port to listen on */
        socklen_t clientlen;
        char *hostaddrp;                /* dotted decimal host addr string */
        int optval;                     /* flag value for setsockopt */
        int n;                          /* message byte size */

        sockfd = socket(AF_INET, SOCK_DGRAM, 0);
        portno = PORT[tid-4];
        optval = 1;
        setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,(const void *)&optval , sizeof(int));
        bzero((char *) &serveraddr, sizeof(serveraddr));
        serveraddr.sin_family = AF_INET;
        serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
        serveraddr.sin_port = htons((unsigned short)portno);
        int b=bind(sockfd, (struct sockaddr *) &serveraddr, sizeof(serveraddr));
        clientlen = sizeof(clientaddr);
        int buf[BUFSIZE];
       
        cout<<"Kernel Thread "<< tid <<" ready to receive..."<<endl;
        while(!TrackDone){
           n = recvfrom(sockfd, buf, BUFSIZE, 0,(struct sockaddr *) &clientaddr, &clientlen);
           POS[tid-4] = Point2f(buf[0], buf[1]); // Retrieve 2D coordinates sent by the camera via UDP
        }
       	
        pthread_exit(NULL);
    }

	

}



// ************************************************************************************ Main Function ***************************************************************************************//
//**************************************************************************************************************************************************************************************//
//**************************************************************************************************************************************************************************************//

void RunEthoLoop( int CAM_NUM, vector<Mat> PM, Mat TransformMatrix, vector<int> AllPorts, int HasLiveCamera, LiveCamera LC,int HasCloseUp, CloseUpSystem CLUP, int HasRecoBox, int WhichRecoBoxMode,ConditioningRecoBox CR,ForagingRecoBox FR,string outputfile)
{

	CameraNumbers=CAM_NUM;
	PORT=AllPorts;
	for(int i=0;i<CameraNumbers;i++){
		projMatrix.push_back(matrixCam*PM[i]);
		POS.push_back(Point2f(0,0));
	}

	RWTransition=TransformMatrix.clone();
	RWTransition.convertTo(RWTransition,CV_32FC1);
	
	UseLiveCamera=HasLiveCamera;
	UseRecoBox=HasRecoBox;
	RecoBoxMode=WhichRecoBoxMode;
	UseCloseUp=HasCloseUp;
	
	LCam=LC;
	FReco=FR;
	CReco=CR;
	CloseUp=CLUP;
	
	OutDir=outputfile;
	OutPutTrajectory.open(OutDir+"/Data.txt");

	TrackDone.store(false);
	gettimeofday(&TBegin,NULL);
	// create the different threads
	int numThreads = 7+ CameraNumbers;
	pthread_t threads[numThreads];
	for(long tt=0; tt<numThreads; tt++){
		usleep(5000);
		printf("In main: creating thread %ld\n", tt);
		int rc = pthread_create(&threads[tt], NULL, Track, (void *)tt);
		if (rc)
		    printf("ERROR in Creating threads..");

	}
	return;
	

}
