/*
 Ali Nouri Zonoz   CMU
 July 2016

 */

#include "opencv2/opencv.hpp"
#include <iomanip>
#include <unistd.h>
#include <arpa/inet.h>
#include <fstream>
#include <atomic>

#define BUFSIZE 	30



using namespace cv;
using namespace std;


extern int Triangulate(const vector<Mat> PM, const vector<int> index, const vector<Point2f> projPoints, Mat& X);

//********************************************************************************************* Defining Global Variables ******************************************************************************************//
//******************************************************************************************************************************************************************************************************************//
//******************************************************************************************************************************************************************************************************************//

int IterNum=0;
std::atomic <bool> CalibDone;

//Integers and others..
int globaltrack=0;						//gnuplot starting
vector<Point2f> pos;					//2D positions received from all cameras
vector<int> portnoList;
int SAMPLE_NUM;
int NUM_CAMS;
int startplot=0;
string FileDirs;
int WorldCoordinateNumber=0;
 // 3D coordinates in real world
vector<Point3f> worldCoord;
// File to write the 3d positions
ofstream myfile3D,myfileRealWorld, Matrix,myfile;                //myfile : get all 3D pos
                                                                //starfile : positions of a 3D star
                                                                //dot : actual 3D position
                                                                //out : file storing the matrices



//*************************************************************************************** Functions **********************************************************************************//
//******************************************************************************************************************************************************************************************//
//*************************************************************************************************************************************************************************************//





/* collectSamplePoints
 * Collects and returns a number of 2D coordinates (SAMPLE_NUM) sent by all cameras
 */
void collectSamplePoints(vector<vector<Point2f> >& pixel_objectPoints, Mat matrixCam, Mat distCoeffs){

    // vector containing 2D coordinates stored regarding the camera number and the sample number
    vector<vector<Point2f> > objectPoints(NUM_CAMS,vector<Point2f>(SAMPLE_NUM));
    int row = 0;
	
    // while we dont have the chosen number of samples
    while(row<SAMPLE_NUM){
        // for each camera, retrieve the 2D coordinates
        for(int i=0; i<NUM_CAMS; i++){
            objectPoints[i][row] = pos[i];
           cout<<objectPoints[i][row].x<<"\t";
        }

        // Only collect points when the data is available from all cameras
        bool notZeros = all_of(objectPoints.begin(), objectPoints.end(), [&row](vector<Point2f> i) { return i[row] != Point2f(0,0); });
        if(notZeros){
        
            cout<<"["<<row<<"]";
            row ++;
            IterNum=IterNum+1;
        }
        //fill(pos.begin(), pos.end(), Point2f(0,0));
		cout<<endl;
        usleep(10000);
        
        
        if(CalibDone) return;
        
    }

    // undistort the 2D points using the corresponding coefficients
    vector<vector<Point2f> > undist_objectPoints(NUM_CAMS,vector<Point2f>(SAMPLE_NUM));
    for(int i=0; i<NUM_CAMS; i++)
        undistortPoints(objectPoints[i], undist_objectPoints[i], matrixCam, distCoeffs);

    // get the 2D coordinates of the projection points in pixel
    pixel_objectPoints.resize(NUM_CAMS,vector<Point2f>(SAMPLE_NUM));
    for (int i=0; i < NUM_CAMS; i++){
        for(int j=0; j < SAMPLE_NUM; j++) {
            pixel_objectPoints[i][j].x = undist_objectPoints[i][j].x* matrixCam.at<float>(0, 0) + matrixCam.at<float>(0, 2);
            pixel_objectPoints[i][j].y = undist_objectPoints[i][j].y* matrixCam.at<float>(1, 1) + matrixCam.at<float>(1, 2);
        }
    }
    for (int i=0; i < NUM_CAMS; i++)
        for(int j=0; j < SAMPLE_NUM; j++)
	    cout<<pixel_objectPoints[i][j].x<<endl;
    cout<<"Collecting points DONE!"<<endl<<endl<<endl;
}

/* getSinglePointCameras
 * get the 2D coordinates (projPoints) and the index of the camera seeing the object
 * only the 2D coordinates where the camera sees the object are taken into account
 */
void getSinglePointCameras(vector<Point2f>& projPoints, vector<int>& index, Mat matrixCam, Mat distCoeffs){

    //Vectors for single points inside the Triangulation loop - for Onlien tracking
    vector<Point2f> obj(NUM_CAMS);

    projPoints.clear();
    index.clear();

    for(int i=0; i<NUM_CAMS; i++)
        obj[i] = pos[i];

    // undistort the 2D points using the corresponding coefficients
    vector<Point2f> undist_obj(NUM_CAMS);
    undistortPoints(obj, undist_obj, matrixCam, distCoeffs);

    // get the 2D coordinates of the projection points in pixel
    vector<Point2f> pixel_obj(NUM_CAMS);
    for(int i=0; i<NUM_CAMS; i++){
        pixel_obj[i].x = undist_obj[i].x * matrixCam.at<float>(0, 0) + matrixCam.at<float>(0, 2);
        pixel_obj[i].y = undist_obj[i].y * matrixCam.at<float>(1, 1) + matrixCam.at<float>(1, 2);
    }

    


    // if the coordinates are (0,0), do not take value
    for (int i=0; i<NUM_CAMS; i++){
        if(obj[i] != Point2f(0,0)){
            index.push_back(i);
            projPoints.push_back(pixel_obj[i]);
        }
    }
}

/* getMatrices
 * get all the extrinsic camera matrices
 */
void getMatrices(vector<Mat>& M,  vector<vector<Point2f> > pixel_objectPoints, Mat matrixCam){

    //-- Camera 1 Matrix
    // The first camera matric is the identity
    M[0] = Mat::eye(3, 4, CV_32FC1);
    cout<<"*** Cam1 matrix"<<endl;
    cout<<M[0]<<endl<<endl;

    //-- Camera 2 Matrix
    // We get the 2nd camera matrix using findEssentialMat and recoverPose
    Mat E1, R1, t1, mask;

    E1 = findEssentialMat(pixel_objectPoints[0], pixel_objectPoints[1], matrixCam, CV_RANSAC, 0.99, 1.0, mask);

    int nb_inliers =  recoverPose(E1, pixel_objectPoints[0], pixel_objectPoints[1], matrixCam, R1, t1,  mask);
    cout << "*** "<< nb_inliers<< " inliers, out of " <<SAMPLE_NUM<<" points" <<endl;

    M[1] = Mat::zeros(3, 4, CV_32FC1);
    R1.copyTo(M[1](Rect(0, 0, 3, 3)));
    t1.copyTo(M[1](Rect(3, 0, 1, 3)));

    cout<<"Cam2 matrix"<<endl;
    cout<<M[1]<<endl<<endl;

    //-- Other Camera Matrices
    Mat point3D, R, t, out_inliers;
    vector<Point3f> pointCloud;

    // Create a 3D set of points with the 2D coordinates of the first two cameras
    triangulatePoints(matrixCam*M[0], matrixCam*M[1], pixel_objectPoints[0], pixel_objectPoints[1], point3D);

    pointCloud.resize( SAMPLE_NUM );
    for (int i = 0; i < pointCloud.size(); i++)
        pointCloud[i] = Point3f(point3D.at<float>(0, i), point3D.at<float>(1, i), point3D.at<float>(2, i)) / point3D.at<float>(3, i);

    // get the camera matrices using the point cloud ad the 2D coordinates of each camera
    for (int i=2; i<NUM_CAMS; i++){
        solvePnPRansac(pointCloud, pixel_objectPoints[i], matrixCam, Mat(), R, t, false, 0.9*SAMPLE_NUM, 2.0,0.99,out_inliers, 0 );
        Rodrigues(R, R);

        cout << "*** " << out_inliers.rows <<" inliers, out of " <<SAMPLE_NUM<<" points" <<endl;

        M[i] = Mat::zeros(3, 4, CV_32FC1);
        R.copyTo(M[i](Rect(0, 0, 3, 3)));
        t.copyTo(M[i](Rect(3, 0, 1, 3)));

        cout<<"Cam"<<i+1<<" matrix"<<endl;
        cout<<M[i]<<endl<<endl;
    }
}






/* Thread function
 * One thread per camera in order to get the object 2D projections coordinates sent by the Jetsons via UDP
 * One thread to plot the triangulated 3D point
 */
void *TrackforCalibration(void *threadid)
{
    long tid;
    tid = (long)threadid;
    
    if(tid==0)
    {
    // variable o wait for the user to press a key to continue
    int press, test;
    //Input: Camera Matrix
    Mat matrixCam   = (Mat_<float>(3,3) << 4.6996008656064788e+02, 0, 3.1550000000000000e+02, 0,  4.6996008656064788e+02, 2.5350000000000000e+02, 0, 0, 1);
    //Input: Distortion Coefficients
    Mat distCoeffs  = (Mat_<float>(1,4) << -3.2842997314171629e-01, 1.6718837850893636e-01, 0, 0);
   	// ******* 1 - Collect points to get extrinsic Matrices *******
    // stores in pixel_objectPoints, the 2D coordinates of the object NUM_SAMPLE times
    vector<vector<Point2f> > pixel_objectPoints;
    collectSamplePoints(pixel_objectPoints, matrixCam, distCoeffs);
	if(CalibDone)
	{pthread_exit(NULL);}
		
    // stoes in M the extrinsic matrices of the cameras
    vector<Mat> M(NUM_CAMS);
    getMatrices(M, pixel_objectPoints, matrixCam );
    // proper camera matrix for Triangulate function is K*[R|t]
    vector<Mat> projMatr(NUM_CAMS);
    for(int i=0; i<NUM_CAMS; i++)
        projMatr[i]=matrixCam*M[i];


	
	Matrix<<"[R_T]"<<endl;
    // stores in the file out the matrices found previously
    for(int k=0;k<NUM_CAMS;k++){
    Matrix<<"M"<<k<<"=";
        for(int i=0;i<3;i++)
            for(int j=0;j<4;j++)
                Matrix<<setprecision(9)<<double(M[k].at<float>(i,j))<<" ";
        Matrix<<endl;
    }
	
    // ******* 2 - TEST NOT REAL WORLD *******
    sleep(1);
   // cout<<"Tracking started..."<<endl;
   // cout<<"press somthing to start"<<endl;
   // cin>>press;

    vector<Point2f> projPoints;
    int count = 0;
    Mat point3D;
    vector<int> index;
   // cout<<"waiting for globalplot"<<endl;
	while(globaltrack==0){}
	
	

    while(globaltrack){
        usleep(3000);
        // get a single 2D coordinates from each camera
        getSinglePointCameras(projPoints, index, matrixCam, distCoeffs);
        // triangulate to get 3D point
        int r = Triangulate(projMatr, index, projPoints, point3D);
        if(r){
           // cout<<"point3D:"<<point3D<<endl;
          //  cout<<count<<endl;
            count++;
            // store the 3D point value
            myfile << point3D.at<float>(0,0)<<" "<<point3D.at<float>(0,1)<<" "<<point3D.at<float>(0,2)<<endl;
            myfile3D << point3D.at<float>(0,0)<<" "<<point3D.at<float>(0,1)<<" "<<point3D.at<float>(0,2)<<endl;
            startplot=1;
             	
        }
        
    }


    // ******* 3 - CREATE MATRICES REAL WORLD *******

    //cout<<"NOW real world calibrationL: press 1"<<endl;
   // cin>>press;
    // 3D computed coordinates
    vector<Point3f> virtualCoord;
    // transformation matrix from computed coordinates to real coordinates
    Mat transfMatrix(3,4,CV_32FC1);

    // stores in virtualCoord the 3D coordinates corresponding to worldCoord
    count = 0;
    while(count<worldCoord.size()){
        cout<<"place your LED at "<<worldCoord[count]<<"  and press number "<<count<<" W is "<<worldCoord.size()<<endl;
        
        while(WorldCoordinateNumber==count){}
        

        usleep(5000);

        getSinglePointCameras(projPoints, index, matrixCam, distCoeffs);

        int r = Triangulate(projMatr, index,projPoints,point3D);
        if(r){
            virtualCoord.push_back(Point3f(point3D.at<float>(0,0),point3D.at<float>(0,1),point3D.at<float>(0,2)));
            count++;
        }
        globaltrack=1;
    }

  //  cout<<worldCoord<<endl;
   // cout<<virtualCoord<<endl;

    // calculates the transformation matrix
    vector<uchar> inliers;
    estimateAffine3D(virtualCoord, worldCoord, transfMatrix, inliers);

    // stores the transformation matrix
    Matrix<<"[RealWorld]"<<endl<<"Real_R_T=";
    for(int i=0;i<3;i++)
        for(int j=0;j<4;j++)
            Matrix<<setprecision(9)<<float(transfMatrix.at<double>(i,j))<<" ";
    Matrix<<endl;

   // cout<<endl<<"This is the matrix"<<endl<<endl;
   // cout<<transfMatrix<<endl;
    
    
	transfMatrix.convertTo(transfMatrix,CV_32FC1);
    // outputs the number of inliers while computing the transformation matrix
   // cout<<"inliers : ";
    for (int i=0;i<worldCoord.size();i++)
        cout<<static_cast<unsigned>(inliers[i]);


    // ******* 4 - TEST REAL WORLD COORD *******


   
   
	globaltrack=0;
	while(globaltrack==0){}
    // real world 3D point
    Mat point3DReal;

    // calculates and plots the 3D coordinates of the object
    while(1){
        usleep(2000);

        getSinglePointCameras(projPoints, index, matrixCam, distCoeffs);

        int r = Triangulate(projMatr, index,projPoints,point3D);
        if(r){
            point3DReal = transfMatrix*point3D;

            myfileRealWorld << point3DReal.at<float>(0,0)<<" "<<point3DReal.at<float>(0,1)<<" "<<point3DReal.at<float>(0,2)<<endl;
            myfile << point3DReal.at<float>(0,0)<<" "<<point3DReal.at<float>(0,1)<<" "<<point3DReal.at<float>(0,2)<<endl;
           if(globaltrack==0)
           	break;
           
        }
    }

}
    
                                    
    if(tid >0 & tid!=NUM_CAMS+1){
        struct sockaddr_in serveraddr;  /* server's addr */
        struct sockaddr_in clientaddr;  /* client addr */
        struct hostent *hostp;          /* client host info */
        int sockfd;                     /* socket */
        int portno;                     /* port to listen on */
        socklen_t clientlen;
        char *hostaddrp;                /* dotted decimal host addr string */
        int optval;                     /* flag value for setsockopt */
        int n;                          /* message byte size */

        sockfd = socket(AF_INET, SOCK_DGRAM, 0);
        portno = portnoList[tid-1];
        optval = 1;
        setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,(const void *)&optval , sizeof(int));
        bzero((char *) &serveraddr, sizeof(serveraddr));
        serveraddr.sin_family = AF_INET;
        serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
        serveraddr.sin_port = htons((unsigned short)portno);
        int b=bind(sockfd, (struct sockaddr *) &serveraddr, sizeof(serveraddr));
        clientlen = sizeof(clientaddr);
        int buf[BUFSIZE];
       
        sleep(1);

        cout<<"Kernel Thread "<< tid <<" ready to receive..."<<endl;
        while(!CalibDone){
           n = recvfrom(sockfd, buf, BUFSIZE, 0,(struct sockaddr *) &clientaddr, &clientlen);
           pos[tid-1] = Point2f(buf[0], buf[1]); // Retrieve 2D coordinates sent by the camera via UDP
         
        }
       
        pthread_exit(NULL);
    }

	if(tid==NUM_CAMS+1)
	{
		string PlottingCommand= "../src/TrackingCalibration/Plot.sh "+FileDirs+" -e";
		while(!CalibDone)
		{
			if(globaltrack==1){
				if (startplot==1)
					system(PlottingCommand.c_str()); 
				}
				
		}
		pthread_exit(NULL);
	}

}

/* createThreads
 * Creates the number of theads needed for all the cameras and the 3D plot
 */
void createThreads(){
    int numThreads = NUM_CAMS + 2;
    pthread_t threads[numThreads];

    for(long tt=0; tt<numThreads; tt++){
        printf("In main: creating thread %ld\n", tt);
        int rc = pthread_create(&threads[tt], NULL, TrackforCalibration, (void *)tt);
        if (rc)
            printf("ERROR in Creating threads..");

    }
}


// **************************************************************************************************** Main Function *************************************************************************************************//
//******************************************************************************************************************************************************************************************************************//
//******************************************************************************************************************************************************************************************************************//

void Calibrate(int Num_CalibPoints,vector <cv::Point3f> RealWorldCoordinates,vector<string> IPstrings,vector<string> JetsonsInfo,vector<int> ports, string SavingDir){

	
	SAMPLE_NUM=Num_CalibPoints;
	 // setup the real coordinates
	worldCoord=RealWorldCoordinates;
	portnoList=ports;
	NUM_CAMS=portnoList.size();
	FileDirs=SavingDir;
	for(int i=0;i<NUM_CAMS;i++)
	{
		pos.push_back(Point2f(0,0));
	}
	
    // create files and folders
    
    
/* createDirectoriesAndFiles
 * Creates different files and directories
*/
	myfile.open(FileDirs+"/DataAll.txt");
    myfile3D.open (FileDirs+"/Data3D.txt");
    myfileRealWorld.open (FileDirs+"/DataReal.txt");
    Matrix.open(FileDirs+"/R_T_Matrix.ini");
    
    Matrix<<"[CamNum]"<<endl<<"NUM="<<NUM_CAMS<<endl<<endl;
    Matrix<<"[Network_Information]"<<endl;
    Matrix<<"Host_IP = "<<IPstrings[0]<<endl;
    Matrix<<"JetsonsName = "<<JetsonsInfo[0]<<endl;
    Matrix<<"JetsonsPass = "<<JetsonsInfo[1]<<endl;
    
    for(int i=1;i<IPstrings.size();i++){
    	Matrix<<"Jetson_"<<i<< "_IP_Port = "<<IPstrings[i]<<"    "<<ports[i-1]<<endl;
    }
    // create the different threads
    createThreads();
}
